"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.array.convolution import convolve_array_simple
from buteo.array.convolution_kernels import _simple_blur_kernel_2d_3x3

EPSILON = 1e-7


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def _get_matched_noise_2d(
    arr: np.ndarray,
    val_min: float = 0.0,
    val_max: float = 1.0,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Matches the noise of an array to the noise of another array.
    """
    size = (1, arr.shape[0], arr.shape[1])
    if channel_last:
        size = (arr.shape[0], arr.shape[1], 1)

    noise = np.random.uniform(val_min, val_max, size=size).astype(np.float32)

    return noise


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def _get_matched_noise_2d_binary(
    arr: np.ndarray,
    val_min: float = 0.0,
    val_max: float = 1.0,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Matches the noise of an array to the noise of another array.
    """
    size = (1, arr.shape[0], arr.shape[1])
    if channel_last:
        size = (arr.shape[0], arr.shape[1], 1)

    binary_noise = np.random.uniform(0.0, 1.0, size=size) > 0.5
    noise = arr.copy()

    _val_min = np.array(val_min, dtype=np.float32)
    _val_max = np.array(val_max, dtype=np.float32)

    if channel_last:
        for col in prange(arr.shape[0]):
            for row in prange(arr.shape[1]):
                if binary_noise[col, row, 0]:
                    noise[col, row, :] = _val_max
                else:
                    noise[col, row, :] = _val_min
    else:
        for col in prange(arr.shape[1]):
            for row in prange(arr.shape[2]):
                if binary_noise[0, col, row]:
                    noise[:, col, row] = _val_max
                else:
                    noise[:, col, row] = _val_min

    return noise


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def _get_matched_noise_3d(
    arr: np.ndarray,
    val_min: float = 0.0,
    val_max: float = 1.0,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Matches the noise of an array to the noise of another array.
    """
    noise = np.random.uniform(val_min, val_max, size=arr.shape).astype(np.float32)

    return noise


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def _get_matched_noise_3d_binary(
    arr: np.ndarray,
    val_min: float = 0.0,
    val_max: float = 1.0,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Matches the noise of an array to the noise of another array.
    """
    binary_noise = np.random.uniform(0.0, 1.0, size=arr.shape) > 0.5
    noise = arr.copy()

    _val_min = np.array(val_min, dtype=np.float32)
    _val_max = np.array(val_max, dtype=np.float32)

    if channel_last:
        for col in prange(arr.shape[0]):
            for row in prange(arr.shape[1]):
                for channel in prange(arr.shape[2]):
                    if binary_noise[col, row, channel]:
                        noise[col, row, channel] = _val_max
                    else:
                        noise[col, row, channel] = _val_min
    else:
        for channel in prange(arr.shape[0]):
            for col in prange(arr.shape[1]):
                for row in prange(arr.shape[2]):
                    if binary_noise[channel, col, row]:
                        noise[channel, col, row] = _val_max
                    else:
                        noise[channel, col, row] = _val_min

    return noise


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def _get_blurred_image(
    X: np.ndarray,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Blurs an image at random. Input should be (height, width, channels) or (channels, height, width).
    
    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.

    Parameters
    ----------
    X : np.ndarray
        The image to blur.

    channel_to_adjust : int, optional
        Weather to only apply the blur to a specific channel.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The blurred image.
    """
    X = X.copy()

    offsets, weights = _simple_blur_kernel_2d_3x3()

    if channel_last:
        for channel in prange(X.shape[2]):
            X[:, :, channel] = convolve_array_simple(
                X[:, :, channel],
                offsets,
                weights
            )
    else:
        for channel in prange(X.shape[0]):
            X[channel, :, :] = convolve_array_simple(
                X[channel, :, :],
                offsets,
                weights,
            )

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def mask_pixels_2d(
    arr: np.ndarray,
    p: float = 0.05,
    channel_last: bool = True,
) -> np.ndarray:
    height, width, channels = arr.shape
    mask_shape = (height, width, 1)

    if not channel_last:
        channels, height, width = arr.shape
        mask_shape = (1, height, width)

    mask = (np.random.uniform(0.0, 1.0, size=mask_shape) > p).astype(np.uint8)
    mask = mask.repeat(channels).reshape(arr.shape)

    return mask


class MaskPixels2D():
    """
    Masks random pixels from an image. The same pixels are masked on each channel.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, channel_last: bool = True):
        self.p = p
        self.channel_last = channel_last

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_pixels_2d(arr, p=self.p, channel_last=self.channel_last)

@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def mask_pixels_3d(
    arr: np.ndarray,
    p: float = 0.05,
    channel_last: bool = True,
) -> np.ndarray:
    mask = (np.random.uniform(0.0, 1.0, size=arr.shape) > p).astype(np.uint8)

    return mask


class MaskPixels3D():
    """
    Masks random pixels from an image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, channel_last: bool = True):
        self.p = p
        self.channel_last = channel_last

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_pixels_3d(arr, p=self.p, channel_last=self.channel_last)

@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def mask_lines_2d(
    arr: np.ndarray,
    p: float = 0.05,
    max_height: float = 1.0,
    max_width: float = 1.0,
    min_height: float = 0.1,
    min_width: float = 0.1,
    max_size: int = 3,
    min_size: int = 1,
    channel_last: bool = True,
):
    height, width, channels = arr.shape
    mask_shape = (height, width, channels)

    if not channel_last:
        channels, height, width = arr.shape
        mask_shape = (channels, height, width)

    mask = np.ones(mask_shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    mask_height = (np.random.uniform(0.0, 1.0, size=(height, )) < p).astype(np.uint8)
    mask_width = (np.random.uniform(0.0, 1.0, size=(width, )) < p).astype(np.uint8)

    mask_height_indices = np.where(mask_height == 1)[0]
    mask_width_indices = np.where(mask_width == 1)[0]

    vertical_lines = mask_height.sum()
    horizontal_lines = mask_width.sum()

    if vertical_lines == 0 and horizontal_lines == 0:
        return mask
    
    vertical_start = np.random.randint(0, height, size=(vertical_lines, ))
    vertical_end = vertical_start + np.floor(np.random.uniform(min_height, max_height, size=(vertical_lines, )) * height).astype(np.int64)

    for i in range(vertical_end.shape[0]):
        if vertical_end[i] > height:
            vertical_end[i] = height

    horizontal_start = np.random.randint(0, width, size=(horizontal_lines, ))
    horizontal_end = horizontal_start + np.floor(np.random.uniform(min_width, max_width, size=(horizontal_lines, )) * width).astype(np.int64)

    for j in range(horizontal_end.shape[0]):
        if horizontal_end[j] > width:
            horizontal_end[j] = width

    if channel_last:
        for idx, pos_y in enumerate(mask_height_indices):
            size = np.random.randint(min_size, max_size + 1)
            size_half = size // 2

            mask[vertical_start[idx]:vertical_end[idx], pos_y - size_half:pos_y + size_half + 1 , :] = zero

        for idx, pos_x in enumerate(mask_width_indices):
            size = np.random.randint(min_size, max_size + 1)
            size_half = size // 2

            mask[pos_x - size_half: pos_x + size_half + 1, horizontal_start[idx]:horizontal_end[idx], :] = zero
    else:
        for idx, pos_y in enumerate(mask_height_indices):
            mask[:, vertical_start[idx]:vertical_end[idx], pos_y + size] = zero

        for idx, pos_x in enumerate(mask_width_indices):
            mask[:, pos_x + size, horizontal_start[idx]:horizontal_end[idx]] = zero

    return mask


class MaskLines2D():
    """
    Masks random lines from an image. The same pixels are masked on each channel.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, channel_last: bool = True):
        self.p = p
        self.channel_last = channel_last

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_lines_2d(arr, p=self.p, channel_last=self.channel_last)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_lines_3d(
    arr: np.ndarray,
    p: float = 0.05,
    channel_last: bool = True,
):

    height, width, channels = arr.shape

    if not channel_last:
        channels, height, width = arr.shape

    mask = np.ones(arr.shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    mask_height = (np.random.uniform(0.0, 1.0, size=(height, channels, )) < p).astype(np.uint8)
    mask_width = (np.random.uniform(0.0, 1.0, size=(width, channels, )) < p).astype(np.uint8)

    if channel_last:
        for col in prange(height):
            for row in prange(width):
                for channel in prange(channels):
                    if mask_height[col, channel]:
                        mask[col, row, channel] = zero
                    if mask_width[row, channel]:
                        mask[col, row, channel] = zero

    else:
        for col in prange(height):
            for row in prange(width):
                for channel in prange(channels):
                    if mask_height[col, channel]:
                        mask[channel, col, row] = zero
                    if mask_width[row, channel]:
                        mask[channel, col, row] = zero

    return mask


class MaskLines3D():
    """
    Masks random lines from an image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, channel_last: bool = True):
        self.p = p
        self.channel_last = channel_last

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_lines_3d(arr, p=self.p, channel_last=self.channel_last)


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def _round_converter(idx, height, width):
    # Top side
    if idx < width:
        return 0, idx

    # Right side
    elif idx < width + height - 1:
        return idx - width + 1, width - 1

    # Bottom side
    elif idx < 2 * width + height - 2:
        return height - 1, width - 1 - (idx - width - height + 2)

    # Left side
    else:
        return height - 1 - (idx - 2 * width - height + 3), 0


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def _quadratic_bezier(px0, py0, px1, py1, px2, py2, t):
    """Calculate point on a quadratic Bezier curve using individual components."""
    a = (1-t)
    b = t
    x = a*(a*px0 + b*px1) + b*(a*px1 + b*px2)
    y = a*(a*py0 + b*py1) + b*(a*py1 + b*py2)
    
    return int(x + 0.5), int(y + 0.5)  # Using +0.5 and int for rounding


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_lines_2d_bezier(
    arr: np.ndarray,
    p: float = 0.05,
    channel_last: bool = True,
) -> np.ndarray:
    if channel_last:
        height, width, channels = arr.shape
    else:
        channels, height, width = arr.shape

    mask = np.ones((height, width, channels), dtype=np.uint8)

    border_pixels = (2 * (height + width)) - 4
    lines_count = (np.random.rand(border_pixels) < p).sum()

    if lines_count == 0:
        return mask

    _sy = np.zeros(lines_count, dtype=np.int64)
    _sx = np.zeros(lines_count, dtype=np.int64)
    _ey = np.zeros(lines_count, dtype=np.int64)
    _ex = np.zeros(lines_count, dtype=np.int64)
    _my = np.zeros(lines_count, dtype=np.int64)
    _mx = np.zeros(lines_count, dtype=np.int64)

    for i in prange(lines_count):
        start_idx = np.random.randint(0, border_pixels)
        end_idx = np.random.randint(0, border_pixels)

        _sy[i], _sx[i] = _round_converter(start_idx, height, width)
        _ey[i], _ex[i] = _round_converter(end_idx, height, width)
        _my[i] = np.random.randint(0, height)
        _mx[i] = np.random.randint(0, width)

    for i in range(lines_count):
        # 1. Calculate the Bounding Box for Each Curve:
        min_x = min(_sx[i], _mx[i], _ex[i])
        max_x = max(_sx[i], _mx[i], _ex[i])
        min_y = min(_sy[i], _my[i], _ey[i])
        max_y = max(_sy[i], _my[i], _ey[i])
        
        # 2. Determine the `linspace` Resolution for Each Curve:
        bounding_box_diagonal = int(np.sqrt((max_x - min_x)**2 + (max_y - min_y)**2))
        space = np.linspace(0.0, 1.0, bounding_box_diagonal)
    
        for t_idx in prange(space.shape[0]):
            t = space[t_idx]
            
            y, x = _quadratic_bezier(_sx[i], _sy[i], _mx[i], _my[i], _ex[i], _ey[i], t)

            if channel_last:
                mask[y, x, :] = 0
            else:
                mask[:, y, x] = 0

    return mask


class MaskLines2DBezier():
    """
    Masks random bezier lines from an image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, channel_last: bool = True):
        self.p = p
        self.channel_last = channel_last

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_lines_2d_bezier(arr, p=self.p, channel_last=self.channel_last)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_lines_3d_bezier(
    arr: np.ndarray,
    p: float = 0.05,
    channel_last: bool = True,
) -> np.ndarray:
    if channel_last:
        height, width, channels = arr.shape
    else:
        channels, height, width = arr.shape

    mask = np.ones((height, width, channels), dtype=np.uint8)

    border_pixels = (2 * (height + width)) - 4

    for c in prange(channels):
        lines_count = (np.random.rand(border_pixels) < p).sum()

        if lines_count == 0:
            continue

        _sy = np.zeros(lines_count, dtype=np.int64)
        _sx = np.zeros(lines_count, dtype=np.int64)
        _ey = np.zeros(lines_count, dtype=np.int64)
        _ex = np.zeros(lines_count, dtype=np.int64)
        _my = np.zeros(lines_count, dtype=np.int64)
        _mx = np.zeros(lines_count, dtype=np.int64)

        for i in range(lines_count):
            start_idx = np.random.randint(0, border_pixels)
            end_idx = np.random.randint(0, border_pixels)

            _sy[i], _sx[i] = _round_converter(start_idx, height, width)
            _ey[i], _ex[i] = _round_converter(end_idx, height, width)
            _my[i] = np.random.randint(0, height)
            _mx[i] = np.random.randint(0, width)

        for i in range(lines_count):
            # 1. Calculate the Bounding Box for Each Curve:
            min_x = min(_sx[i], _mx[i], _ex[i])
            max_x = max(_sx[i], _mx[i], _ex[i])
            min_y = min(_sy[i], _my[i], _ey[i])
            max_y = max(_sy[i], _my[i], _ey[i])
            
            # 2. Determine the `linspace` Resolution for Each Curve:
            bounding_box_diagonal = int(np.sqrt((max_x - min_x)**2 + (max_y - min_y)**2))
            space = np.linspace(0.0, 1.0, bounding_box_diagonal)
        
            for t_idx in prange(space.shape[0]):
                t = space[t_idx]
                
                y, x = _quadratic_bezier(_sx[i], _sy[i], _mx[i], _my[i], _ex[i], _ey[i], t)

                if channel_last:
                    mask[y, x, c] = 0
                else:
                    mask[c, y, x] = 0

    return mask


class MaskLines3DBezier():
    """
    Masks random bezier lines from an image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, channel_last: bool = True):
        self.p = p
        self.channel_last = channel_last

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_lines_3d_bezier(arr, p=self.p, channel_last=self.channel_last)


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def _point_within_elipse(
    centroid: np.ndarray,
    a: float,
    b: float,
    theta: float,
    point: np.ndarray,
) -> bool:
    x, y = point
    x0, y0 = centroid

    return ((x - x0) * np.cos(theta) + (y - y0) * np.sin(theta)) ** 2 / a ** 2 + ((x - x0) * np.sin(theta) - (y - y0) * np.cos(theta)) ** 2 / b ** 2 <= 1


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_elipse_2d(
    arr: np.ndarray,
    p: float = 0.05,
    max_height: float = 0.4,
    max_width: float = 0.4,
    min_height: float = 0.1,
    min_width: float = 0.1,
    channel_last: bool = True,
):
    height, width, channels = arr.shape
    mask_shape = (height, width, channels)

    if not channel_last:
        channels, height, width = arr.shape
        mask_shape = (channels, height, width)

    mask = np.ones(mask_shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    if np.random.uniform(0.0, 1.0 + EPSILON) > p:
        return mask

    # Randomly select the centroid of the ellipse
    centroid = (np.random.uniform(0, height), np.random.uniform(0, width))
    
    # Randomly select the height and width of the ellipse
    # Ensure that the ellipse fits within the rectangle
    min_x = int(width * min_width)
    min_y = int(height * min_height)
    max_x = int(width * max_width)
    max_y = int(height * max_height)
    a = np.random.uniform(min_y, max_y)  # semi-major axis
    b = np.random.uniform(min_x, max_x)  # semi-minor axis
    
    # Randomly select the orientation of the ellipse
    theta = np.random.uniform(0, 2 * np.pi)

    # Calculate lazy bounds
    max_side = np.ceil(np.maximum(a, b))
    max_y = centroid[0] + max_side
    min_y = centroid[0] - max_side
    max_x = centroid[1] + max_side
    min_x = centroid[1] - max_side

    for col in prange(height):
        if col < min_y or col > max_y:
            continue
        for row in prange(width):
            if row < min_x or row > max_x:
                continue

            if _point_within_elipse(centroid, a, b, theta, np.array([col, row], dtype=np.float32)):
                if channel_last:
                    mask[col, row, :] = zero
                else:
                    mask[:, col, row] = zero

    return mask


class MaskElipse2D():
    """
    Masks a random elipse within the image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self,
        p: float = 0.05,
        channel_last: bool = True,
        max_height: float = 0.4,
        max_width: float = 0.4,
        min_height: float = 0.1,
        min_width: float = 0.1,
    ):
        self.p = p
        self.channel_last = channel_last
        self.max_height = max_height
        self.max_width = max_width
        self.min_height = min_height
        self.min_width = min_width

        assert self.max_height >= self.min_height, "max_height must be greater than or equal to min_height"
        assert self.max_width >= self.min_width, "max_width must be greater than or equal to min_width"
        assert p >= 0.0 and p <= 1.0, "p must be between 0.0 and 1.0"
        assert max_height >= 0.0 and max_height <= 1.0, "max_height must be between 0.0 and 1.0"
        assert max_width >= 0.0 and max_width <= 1.0, "max_width must be between 0.0 and 1.0"
        assert min_height >= 0.0 and min_height <= 1.0, "min_height must be between 0.0 and 1.0"
        assert min_width >= 0.0 and min_width <= 1.0, "min_width must be between 0.0 and 1.0"

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_elipse_2d(
            arr,
            p=self.p,
            max_height=self.max_height,
            max_width=self.max_width,
            min_height=self.min_height,
            min_width=self.min_width,
            channel_last=self.channel_last,
        )


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_elipse_3d(
    arr: np.ndarray,
    p: float = 0.05,
    max_height: float = 0.4,
    max_width: float = 0.4,
    min_height: float = 0.1,
    min_width: float = 0.1,
    channel_last: bool = True,
):
    height, width, channels = arr.shape
    mask_shape = (height, width, channels)

    if not channel_last:
        channels, height, width = arr.shape
        mask_shape = (channels, height, width)

    mask = np.ones(mask_shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    for c in range(channels):
        if np.random.uniform(0.0, 1.0 + EPSILON) > p:
            continue

        # Randomly select the centroid of the ellipse
        centroid = (np.random.uniform(0, height), np.random.uniform(0, width))
        
        # Randomly select the height and width of the ellipse
        # Ensure that the ellipse fits within the rectangle
        min_x = int(width * min_width)
        min_y = int(height * min_height)
        max_x = int(width * max_width)
        max_y = int(height * max_height)
        a = np.random.uniform(min_y, max_y)  # semi-major axis
        b = np.random.uniform(min_x, max_x)  # semi-minor axis
        
        # Randomly select the orientation of the ellipse
        theta = np.random.uniform(0, 2 * np.pi)

        # Calculate lazy bounds
        max_side = np.ceil(np.maximum(a, b))
        max_y = centroid[0] + max_side
        min_y = centroid[0] - max_side
        max_x = centroid[1] + max_side
        min_x = centroid[1] - max_side

        for col in prange(height):
            if col < min_y or col > max_y:
                continue
            for row in prange(width):
                if row < min_x or row > max_x:
                    continue

                if _point_within_elipse(centroid, a, b, theta, np.array([col, row], dtype=np.float32)):
                    if channel_last:
                        mask[col, row, c] = zero
                    else:
                        mask[c, col, row] = zero

    return mask


class MaskElipse3D():
    """
    Masks a random elipse within the image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self,
        p: float = 0.05,
        max_height: float = 0.4,
        max_width: float = 0.4,
        min_height: float = 0.1,
        min_width: float = 0.1,
        channel_last: bool = True,
    ):
        self.p = p
        self.channel_last = channel_last
        self.max_height = max_height
        self.max_width = max_width
        self.min_height = min_height
        self.min_width = min_width

        assert self.max_height >= self.min_height, "max_height must be greater than or equal to min_height"
        assert self.max_width >= self.min_width, "max_width must be greater than or equal to min_width"
        assert p >= 0.0 and p <= 1.0, "p must be between 0.0 and 1.0"
        assert max_height >= 0.0 and max_height <= 1.0, "max_height must be between 0.0 and 1.0"
        assert max_width >= 0.0 and max_width <= 1.0, "max_width must be between 0.0 and 1.0"
        assert min_height >= 0.0 and min_height <= 1.0, "min_height must be between 0.0 and 1.0"
        assert min_width >= 0.0 and min_width <= 1.0, "min_width must be between 0.0 and 1.0"

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_elipse_3d(
            arr,
            p=self.p,
            channel_last=self.channel_last,
            max_height=self.max_height,
            max_width=self.max_width,
            min_height=self.min_height,
            min_width=self.min_width,
        )


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_rectangle_2d(
    arr: np.ndarray,
    p: float = 0.05,
    max_height: float = 0.5,
    max_width: float = 0.5,
    min_height: float = 0.1,
    min_width: float = 0.1,
    channel_last: bool = True,
) -> np.ndarray:
    
    height, width, _channels = arr.shape

    if not channel_last:
        _channels, height, width = arr.shape

    max_height = int(max_height * height)
    max_width = int(max_width * width)
    min_height = int(min_height * height)
    min_width = int(min_width * width)

    mask = np.ones(arr.shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    if np.random.uniform(0.0, 1.0 + EPSILON) > p:
        return mask

    if max_height == -1:
        mask_height = np.random.randint(min_height, (height // 2) + 1)
    else:
        mask_height = np.random.randint(min_height, max_height + 1)
    
    if max_width == -1:
        mask_width = np.random.randint(min_width, (width // 2) + 1)
    else:
        mask_width = np.random.randint(min_width, max_width + 1)

    col = np.random.randint(0, height - mask_height)
    row = np.random.randint(0, width - mask_width)

    if channel_last:
        mask[col:col + mask_height, row:row + mask_width, :] = zero
    else:
        mask[:, col:col + mask_height, row:row + mask_width] = zero

    return mask


class MaskRectangle2D():
    """
    Masks a random rectangle from an image. The same pixels are masked on each channel.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of a rectangle being masked, default: 0.05.

    max_height : int, optional
        The maximum height (proportion of total_height) of the rectangle, default: 0.5.

    max_width : int, optional
        The maximum width (proportion of total_width) of the rectangle, default: 0.5.

    min_height : int, optional
        The minimum height of the rectangle (proportion of total_height), default: 0.1

    min_width : int, optional
        The minimum width of the rectangle (proportion of total_width), default: 0.1

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(
        self,
        p: float = 0.05,
        max_height: float = 0.5,
        max_width: float = 0.5,
        min_height: float = 0.1,
        min_width: float = 0.1,
        channel_last: bool = True,
    ):
        self.p = p
        self.max_height = max_height
        self.max_width = max_width
        self.min_height = min_height
        self.min_width = min_width
        self.channel_last = channel_last

        assert self.max_height >= self.min_height, "max_height must be greater than or equal to min_height"
        assert self.max_width >= self.min_width, "max_width must be greater than or equal to min_width"
        assert p >= 0.0 and p <= 1.0, "p must be between 0.0 and 1.0"
        assert max_height >= 0.0 and max_height <= 1.0, "max_height must be between 0.0 and 1.0"
        assert max_width >= 0.0 and max_width <= 1.0, "max_width must be between 0.0 and 1.0"
        assert min_height >= 0.0 and min_height <= 1.0, "min_height must be between 0.0 and 1.0"
        assert min_width >= 0.0 and min_width <= 1.0, "min_width must be between 0.0 and 1.0"

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_rectangle_2d(
            arr,
            p=self.p,
            max_height=self.max_height,
            max_width=self.max_width,
            min_height=self.min_height,
            min_width=self.min_width,
            channel_last=self.channel_last,
        )
    

@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_rectangle_3d(
    arr: np.ndarray,
    p: float = 0.05,
    max_height: float = 0.5,
    max_width: float = 0.5,
    min_height: float = 0.1,
    min_width: float = 0.1,
    channel_last: bool = True,
) -> np.ndarray:
    height, width, channels = arr.shape

    if not channel_last:
        channels, height, width = arr.shape

    mask = np.ones(arr.shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    max_height = int(max_height * height)
    max_width = int(max_width * width)
    min_height = int(min_height * height)
    min_width = int(min_width * width)

    for c in prange(channels):
        if np.random.uniform(0.0, 1.0 + EPSILON) > p:
            continue

        if max_height == -1:
            mask_height = np.random.randint(min_height, (height // 2) + 1)
        else:
            mask_height = np.random.randint(min_height, max_height + 1)
        
        if max_width == -1:
            mask_width = np.random.randint(min_width, (width // 2) + 1)
        else:
            mask_width = np.random.randint(min_width, max_width + 1)

        col = np.random.randint(0, height - mask_height)
        row = np.random.randint(0, width - mask_width)

        if channel_last:
            mask[col:col + mask_height, row:row + mask_width, c] = zero
        else:
            mask[c, col:col + mask_height, row:row + mask_width] = zero

    return mask


class MaskRectangle3D():
    """
    Masks a random rectangle from an image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of a rectangle being masked, default: 0.05.

    max_height : int, optional
        The maximum height (proportion of total_height) of the rectangle, default: 0.5.

    max_width : int, optional
        The maximum width (proportion of total_width) of the rectangle, default: 0.5.

    min_height : int, optional
        The minimum height of the rectangle (proportion of total_height), default: 0.1

    min_width : int, optional
        The minimum width of the rectangle (proportion of total_width), default: 0.1

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(
        self,
        p: float = 0.05,
        max_height: float = 0.5,
        max_width: float = 0.5,
        min_height: float = 0.1,
        min_width: float = 0.1,
        channel_last: bool = True,
    ):
        self.p = p
        self.max_height = max_height
        self.max_width = max_width
        self.min_height = min_height
        self.min_width = min_width
        self.channel_last = channel_last

        assert self.max_height >= self.min_height, "max_height must be greater than or equal to min_height"
        assert self.max_width >= self.min_width, "max_width must be greater than or equal to min_width"
        assert p >= 0.0 and p <= 1.0, "p must be between 0.0 and 1.0"
        assert max_height >= 0.0 and max_height <= 1.0, "max_height must be between 0.0 and 1.0"
        assert max_width >= 0.0 and max_width <= 1.0, "max_width must be between 0.0 and 1.0"
        assert min_height >= 0.0 and min_height <= 1.0, "min_height must be between 0.0 and 1.0"
        assert min_width >= 0.0 and min_width <= 1.0, "min_width must be between 0.0 and 1.0"

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_rectangle_3d(
            arr,
            p=self.p,
            max_height=self.max_height,
            max_width=self.max_width,
            min_height=self.min_height,
            min_width=self.min_width,
            channel_last=self.channel_last,
        )


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def mask_channels(
    arr: np.ndarray,
    p: float = 0.05,
    channel_last: bool = True,
    max_channels: int = 1,
) -> np.ndarray:
    _height, _width, channels = arr.shape

    if not channel_last:
        channels, _height, _width = arr.shape

    mask = np.ones(arr.shape, dtype=np.uint8)
    zero = np.array(0, dtype=np.uint8)

    for _ in prange(max_channels):
        if np.random.uniform(0.0, 1.0) < p:
            random_channel = np.random.randint(0, channels)

            if channel_last:
                mask[:, :, random_channel] = zero
            else:
                mask[random_channel, :, :] = zero

    return mask


class MaskChannels():
    """
    Masks random channels from an image.

    Parameters
    ----------
    arr : np.ndarray
        The image to mask a pixel from.

    p : float, optional
        The probability of masking a pixel, default: 0.05.

    max_channels : int, optional
        The maximum number of channels to mask, default: 1.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The image with masked pixels.
    """
    def __init__(self, p: float = 0.05, max_channels: int = 1, channel_last: bool = True):
        self.p = p
        self.max_channels = max_channels
        self.channel_last = channel_last

        assert p >= 0.0 and p <= 1.0, "p must be between 0.0 and 1.0"
        if channel_last:
            assert max_channels >= 1, "max_channels must be between 1 and the number of channels in the image"

    def __call__(self, arr: np.ndarray) -> np.ndarray:
        return mask_channels(arr, p=self.p, max_channels=self.max_channels, channel_last=self.channel_last)


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def mask_replace_2d(
    arr: np.ndarray,
    mask: np.ndarray,
    method: int = 0,
    val: float = 0.0,
    min_val: float = 0.0,
    max_val: float = 1.0,
    channel_last: bool = True,
    inplace: bool = False
):
    """
    Replaces pixels in an array with values using a mask and a method.
    NOTE: The mask should have the same shape as the array and
    contain values either 0 for masked and 1 for valid.

    Parameters
    ----------
    arr : np.ndarray
        The array to replace pixels in.

    mask : np.ndarray
        The mask to use for replacing pixels.

    method : int, optional
        The method to use for replacing pixels.
        0. replace with 0
        1. replace with 1
        2. replace with val
        3. replace with random value between min_val and max_val
        4. replace with random value binary value, either min_val or max_val
        5. replace with a blurred version of the image

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The array with replaced pixels.
    """
    replace_arr_needed = False
    replace_val = 0.0
    replace_arr = None

    if inplace:
        X = arr
    else:
        X = arr.copy()

    if method == 0:
        replace_arr_needed = False
        replace_val = 0.0
    elif method == 1:
        replace_arr_needed = False
        replace_val = 1.0
    elif method == 2:
        replace_arr_needed = False
        replace_val = val
    elif method == 3:
        replace_arr_needed = True
        replace_val = 0.0
        replace_arr = _get_matched_noise_2d(X, min_val, max_val, channel_last)
    elif method == 4:
        replace_arr_needed = True
        replace_val = 0.0
        replace_arr = _get_matched_noise_2d_binary(X, min_val, max_val, channel_last)
    elif method == 5:
        replace_arr_needed = True
        replace_val = 0.0
        replace_arr = _get_blurred_image(X, channel_last)
    else:
        raise ValueError("method must be between 0 and 4.")
    
    if channel_last:
        for col in prange(X.shape[0]):
            for row in prange(X.shape[1]):
                if mask[col, row, 0] == 0:
                    if replace_arr_needed:
                        X[col, row, :] = replace_arr[col, row, :]
                    else:
                        X[col, row, :] = replace_val
    else:
        for col in prange(X.shape[1]):
            for row in prange(X.shape[2]):
                if mask[0, col, row] == 0:
                    if replace_arr_needed:
                        X[:, col, row] = replace_arr[:, col, row]
                    else:
                        X[:, col, row] = replace_val

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def mask_replace_3d(
    arr: np.ndarray,
    mask: np.ndarray,
    method: int = 0,
    val: float = 0.0,
    min_val: float = 0.0,
    max_val: float = 1.0,
    channel_last: bool = True,
    inplace: bool = False
):
    """
    Replaces pixels in an array with values using a mask and a method.
    NOTE: The mask should have the same shape as the array and
    contain values either 0 for masked and 1 for valid.

    Parameters
    ----------
    arr : np.ndarray
        The array to replace pixels in.

    mask : np.ndarray
        The mask to use for replacing pixels.

    method : int, optional
        The method to use for replacing pixels.
        0. replace with 0
        1. replace with 1
        2. replace with val
        3. replace with random value between min_val and max_val
        4. replace with random value binary value, either min_val or max_val
        5. replace with a blurred version of the image

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    Returns
    -------
    np.ndarray
        The array with replaced pixels.
    """
    replace_arr_needed = False
    replace_val = 0.0
    replace_arr = None

    if inplace:
        X = arr
    else:
        X = arr.copy()

    if method == 0:
        replace_arr_needed = False
        replace_val = 0.0
    elif method == 1:
        replace_arr_needed = False
        replace_val = 1.0
    elif method == 2:
        replace_arr_needed = False
        replace_val = val
    elif method == 3:
        replace_arr_needed = True
        replace_val = 0.0
        replace_arr = _get_matched_noise_3d(X, min_val, max_val, channel_last)
    elif method == 4:
        replace_arr_needed = True
        replace_val = 0.0
        replace_arr = _get_matched_noise_3d_binary(X, min_val, max_val, channel_last)
    elif method == 5:
        replace_arr_needed = True
        replace_val = 0.0
        replace_arr = _get_blurred_image(X, channel_last)
    else:
        raise ValueError("method must be between 0 and 4.")
    
    if channel_last:
        for col in prange(X.shape[0]):
            for row in prange(X.shape[1]):
                for channel in prange(X.shape[2]):
                    if mask[col, row, channel] == 0:
                        if replace_arr_needed:
                            X[col, row, channel] = replace_arr[col, row, channel]
                        else:
                            X[col, row, channel] = replace_val
    else:
        for channel in prange(X.shape[0]):
            for col in prange(X.shape[1]):
                for row in prange(X.shape[2]):
                    if mask[channel, col, row] == 0:
                        if replace_arr_needed:
                            X[channel, col, row] = replace_arr[channel, col, row]
                        else:
                            X[channel, col, row] = replace_val

    return X



"""
# Geospatial Analysis Meets AI

Buteo is a toolbox designed to simplify the process of working with geospatial data in machine learning. It includes tools for reading, writing, and processing geospatial data, as well as tools for creating labels from vector data and generating patches from geospatial data. Buteo makes it easy to ingest data, create training data, and perform inference on geospatial data.

Please note that Buteo is under active development, and its API may not be entirely stable.
When using, please pin the version of Buteo you are using to avoid breaking changes.
Feel free to report any bugs or suggest improvements.

Author: *Casper Fibaek* (github.com/casperfibaek) </br>

**Dependencies** </br>
`numba` (https://numba.pydata.org/) </br>
`gdal` (https://gdal.org/) </br>

**Installation** </br>
Using pip:
```
pip install gdal
pip install buteo
```
Using conda:
```
conda install gdal
pip install buteo
```

**Quickstart**

### Reproject (and other functions) to references. (Vector and raster)
```python
import buteo as beo

OUTDIR = "path/to/output/dir"

vector_file_correct_projection = "path/to/vector/file.gpkg"
raster_files_wrong_projection = "path/to/raster/files/*.tif:glob"

paths_to_reprojected_rasters = beo.reproject_raster(
    raster_files_with_wrong_projection,
    vector_file_with_correct_projection,
    out_path=outdir
)

paths_to_reprojected_rasters
>>> [path/to/output/dir/file1.tif, path/to/output/dir/file2.tif, ...]
```

### Align, stack, and make patches from rasters
```python
import buteo as beo

SRCDIR = "path/to/src/dir/"

paths_to_aligned_rasters_in_memory = beo.align_rasters(
    SRCDIR + "*.tif:glob",
)

stacked_numpy_arrays = beo.raster_to_array(
    paths_to_aligned_rasters_in_memory,
)

patches = beo.array_to_patches(
    stacked_numpy_arrays,
    256,
    n_offsets=1, # 1 overlap at 1/2 patch size (128)
)

# patches_nr, height, width, channels
patches
>>> np.ndarray([10000, 256, 256, 9])
```

### Predict a raster using a model
```python
import buteo as beo

RASTER_PATH = "path/to/raster/raster.tif"
RASTER_OUT_PATH = "path/to/raster/raster_pred.tif"

array = beo.raster_to_array(RASTER_PATH)

callback = model.predict # from pytorch, keras, etc..

# Predict the raster using overlaps, and borders.
# Merge using different methods. (median, mad, mean, mode, ...)
predicted = predict_array(
    array,
    callback,
    tile_size=256,
)

# Write the predicted raster to disk
beo.array_to_raster(
    predicted,
    reference=RASTER_PATH,
    out_path=RASTER_OUT_PATH,
)
# Path to the predicted raster
>>> "path/to/raster/raster_pred.tif"
```

</br>

| Example Colabs                        |                                                                                                                                                                                                               |
|---------------------------------------|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Create labels from OpenStreetMap data | [![Open All Collab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/casperfibaek/buteo/blob/master/examples/create_labels_from_osm_data.ipynb)            |
| Scheduled cleaning of geospatial data | [![Open All Collab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/casperfibaek/buteo/blob/master/examples/rocket_example.ipynb)                         |
| Clip and remove noise from rasters    | [![Open All Collab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/casperfibaek/buteo/blob/master/examples/clip_and_remove_noise_raster.ipynb)           |
| Sharpen nightlights data              | [![Open All Collab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/casperfibaek/buteo/blob/master/examples/process_nightlights.ipynb)                    |
| Filters and morphological operations  | [![Open All Collab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/casperfibaek/buteo/blob/master/examples/filters_and_morphology.ipynb)                 |


</br>

The toolbox is being developed by ESA-Philab, NIRAS, and Aalborg University.

# Dependencies
gdal
numba

optional:
orfeo-toolbox
esa-snap
"""
from .utils import *
from .raster import *
from .vector import *
from .array import *
from .eo import *
from .ai import *

try:
    from osgeo import gdal
    gdal.UseExceptions()
except:
    print("GDAL not installed. Some functions may not work.")

__version__ = "0.9.53"


"""
### Bounding box utility functions ###

Various utility functions to work with bounding boxes and gdal.

There are two different formats for bounding boxes used by GDAL:</br>
OGR:  `[x_min, x_max, y_min, y_max]`</br>
WARP: `[x_min, y_min, x_max, y_max]`</br>

_If nothing else is stated, the OGR format is used._

The GDAL geotransform is a list of six parameters:</br>
`x_min, pixel_width, row_skew, y_max, column_skew, pixel_height (negative for north-up)`
"""

# Standard library
import sys; sys.path.append("../../")
from typing import List, Union, Dict, Any, Optional
from uuid import uuid4

# External
import numpy as np
from osgeo import ogr, osr, gdal

# Internal
from buteo.utils import utils_base, utils_projection, utils_path, utils_gdal



def _check_is_valid_bbox(bbox_ogr: List[Union[int, float]]) -> bool:
    """
    Checks if a bbox is valid.

    A valid ogr formatted bbox has the form:
        `[x_min, x_max, y_min, y_max]`

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    Returns
    -------
    bool
        True if the bbox is valid, False otherwise.
    """
    if not isinstance(bbox_ogr, list):
        return False

    if len(bbox_ogr) != 4:
        return False

    for val in bbox_ogr:
        if not utils_base._check_variable_is_number_type(val):
            return False

    x_min, x_max, y_min, y_max = bbox_ogr

    if x_min > x_max or y_min > y_max:
        return False

    if True in [np.isinf(val) for val in bbox_ogr]:
        return True

    return True


def _check_is_valid_bbox_latlng(bbox_ogr_latlng: List[Union[int, float]]) -> bool:
    """
    Checks if a bbox is valid and latlng.

    A valid ogr formatted bbox has the form:
        `[x_min, x_max, y_min, y_max]`

    Parameters
    ----------
    bbox_ogr_latlng : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
        
    Returns
    -------
    bool
        True if the bbox is valid, False otherwise.
    """
    if not _check_is_valid_bbox(bbox_ogr_latlng):
        return False

    x_min, x_max, y_min, y_max = bbox_ogr_latlng
    if x_min < -180 or x_max > 180 or y_min < -90 or y_max > 90:
        return False

    return True


def _check_is_valid_geotransform(geotransform: List[Union[int, float]]) -> bool:
    """
    Checks if a geotransform is valid.

    A valid geotransform has the form:
    `[x_min, pixel_width, row_skew, y_max, column_skew, pixel_height]`

    Parameters
    ----------
    geotransform : list
        A GDAL formatted geotransform.
        `[x_min, pixel_width, row_skew, y_max, column_skew, pixel_height]`
        pixel_height is negative for north up.

    Returns
    -------
    bool
        True if the geotransform is valid, False otherwise.
    """
    if not isinstance(geotransform, (list, tuple)):
        return False

    if len(geotransform) != 6:
        return False

    for val in geotransform:
        if not utils_base._check_variable_is_number_type(val):
            return False

    return True


def _get_pixel_offsets(
    geotransform: List[Union[int, float]],
    bbox_ogr: List[Union[int, float]],
):
    """
    Get the pixels offsets for a bbox and a geotransform.

    Parameters
    ----------
    geotransform : list
        A GDAL formatted geotransform.
        `[x_min, pixel_width, row_skew, y_max, column_skew, pixel_height]`
        pixel_height is negative for north up.

    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    Returns
    -------
    list
        A list of pixel offsets. `[x_start, y_start, x_size, y_size]`
    """
    assert _check_is_valid_geotransform(
        geotransform
    ), f"geotransform must be a list of length six. Received: {geotransform}."
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr must be a valid OGR formatted bbox. Received: {bbox_ogr}."

    x_min, x_max, y_min, y_max = bbox_ogr

    origin_x = geotransform[0]
    origin_y = geotransform[3]
    pixel_width = geotransform[1]
    pixel_height = geotransform[5]

    x_start = int(np.rint((x_min - origin_x) / pixel_width))
    y_start = int(np.rint((y_max - origin_y) / pixel_height))
    x_size = int(np.rint((x_max - x_min) / pixel_width))
    y_size = int(np.rint((y_min - y_max) / pixel_height))

    return [x_start, y_start, x_size, y_size]


def _get_bbox_from_geotransform(
    geotransform: List[Union[int, float]],
    raster_x_size: int,
    raster_y_size: int,
) -> List[Union[int, float]]:
    """
    Get an OGR bounding box from a geotransform and raster sizes.

    Parameters
    ----------
    geotransform : list
        A GDAL formatted geotransform.
        `[x_min, pixel_width, row_skew, y_max, column_skew, pixel_height]`

    raster_x_size : int
        The number of pixels in the x direction.

    raster_y_size : int
        The number of pixels in the y direction.

    Returns
    -------
    list
        An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert _check_is_valid_geotransform(
        geotransform
    ), f"geotransform was not a valid geotransform. Received: {geotransform}"

    x_min, pixel_width, _row_skew, y_max, _column_skew, pixel_height = geotransform

    x_max = x_min + (raster_x_size * pixel_width)
    y_min = y_max + (raster_y_size * pixel_height)

    return [x_min, x_max, y_min, y_max]


def _get_bbox_from_raster(
    raster_dataframe: gdal.Dataset,
) -> List[Union[int, float]]:
    """
    Gets an OGR bounding box from a GDAL raster dataframe.
    The bounding box is in the same projection as the raster.

    Parameters
    ----------
    raster_dataframe : gdal.Dataset
        A GDAL raster dataframe.

    Returns
    -------
    list
        An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        raster_dataframe, gdal.Dataset
    ), f"raster_dataframe was not a gdal.Datasource. Received: {raster_dataframe}"

    bbox = _get_bbox_from_geotransform(
        raster_dataframe.GetGeoTransform(),
        raster_dataframe.RasterXSize,
        raster_dataframe.RasterYSize,
    )

    return bbox


def _get_bbox_from_vector(
    vector_dataframe: ogr.DataSource,
) -> List[Union[int, float]]:
    """
    Gets an OGR bounding box from an OGR dataframe.
    The bounding box is in the same projection as the vector.

    Parameters
    ----------
    vector_dataframe : ogr.DataSource
        An OGR vector dataframe.

    Returns
    -------
    list
        An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        vector_dataframe, ogr.DataSource
    ), f"vector_dataframe was not a valid ogr.DataSource. Received: {vector_dataframe}"

    layer_count = vector_dataframe.GetLayerCount()

    assert (
        layer_count > 0
    ), f"vector_dataframe did not contain any layers. Received: {vector_dataframe}"

    x_min = None
    x_max = None
    y_min = None
    y_max = None

    for layer_index in range(layer_count):
        layer = vector_dataframe.GetLayerByIndex(layer_index)
        layer_x_min, layer_x_max, layer_y_min, layer_y_max = layer.GetExtent()

        if layer_index == 0:
            x_min = layer_x_min
            x_max = layer_x_max
            y_min = layer_y_min
            y_max = layer_y_max
        else:
            if layer_x_min < x_min:
                x_min = layer_x_min
            if layer_x_max > x_max:
                x_max = layer_x_max
            if layer_y_min < y_min:
                y_min = layer_y_min
            if layer_y_max > y_max:
                y_max = layer_y_max

    return [x_min, x_max, y_min, y_max]


def _get_bbox_from_vector_layer(
    vector_layer: ogr.Layer,
) -> List[Union[int, float]]:
    """
    Gets an OGR bounding box from an OGR dataframe layer.

    Parameters
    ----------
    vector_layer : ogr.Layer
        An OGR vector dataframe layer.

    Returns
    -------
    list
        An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        vector_layer, ogr.Layer
    ), f"vector_layer was not a valid ogr.Layer. Received: {vector_layer}"

    x_min, x_max, y_min, y_max = vector_layer.GetExtent()

    return [x_min, x_max, y_min, y_max]


def get_bbox_from_dataset(
    dataset: Union[str, gdal.Dataset, ogr.DataSource],
) -> List[Union[int, float]]:
    """
    Get the bbox from a dataset.
    The bounding box is in the same projection as the dataset.
    If you want the Bounding Box in WGS84, use `get_bbox_from_dataset_wgs84`.

    Parameters
    ----------
    dataset : str, gdal.Dataset, ogr.DataSource
        A dataset or dataset path.

    Returns
    -------
    list
        The bounding box in ogr format: `[x_min, x_max, y_min, y_max]`.
    """
    assert isinstance(dataset, (str, gdal.Dataset, ogr.DataSource)), "DataSet must be a string, ogr.DataSource, or gdal.Dataset."

    opened = dataset if isinstance(dataset, (gdal.Dataset, ogr.DataSource)) else None

    if opened is None:
        gdal.PushErrorHandler("CPLQuietErrorHandler")
        opened = gdal.Open(dataset, gdal.GA_ReadOnly)

        if opened is None:
            opened = ogr.Open(dataset, gdal.GA_ReadOnly)

        gdal.PopErrorHandler()
        if opened is None:
            raise RuntimeError(f"Could not open dataset. {dataset}")

    if isinstance(opened, gdal.Dataset):
        return _get_bbox_from_raster(opened)

    if isinstance(opened, ogr.DataSource):
        return _get_bbox_from_vector(opened)

    raise RuntimeError(f"Could not get bbox from dataset. {dataset}")


def _get_sub_geotransform(
    geotransform: List[Union[int, float]],
    bbox_ogr: List[Union[int, float]],
) -> Dict:
    """
    Create a GeoTransform and the raster sizes for an OGR formatted bbox.

    Parameters
    ----------
    geotransform : list
        A GDAL geotransform.
        `[top_left_x, pixel_width, rotation_x, top_left_y, rotation_y, pixel_height]`
    
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    Returns
    -------
    dict
        { "Transform": _list_, "RasterXSize": _int_, "RasterYSize": _int_ }
    """
    assert _check_is_valid_geotransform(
        geotransform
    ), f"geotransform must be a valid geotransform. Received: {geotransform}."
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr
    pixel_width = geotransform[1]
    pixel_height = geotransform[5]

    raster_x_size = round((x_max - x_min) / pixel_width)
    raster_y_size = round((y_max - y_min) / pixel_height)

    return {
        "Transform": [x_min, pixel_width, 0, y_max, 0, utils_base._ensure_negative(pixel_height)],
        "RasterXSize": abs(raster_x_size),
        "RasterYSize": abs(raster_y_size),
    }


def _get_geom_from_bbox(
    bbox_ogr: List[Union[int, float]],
) -> ogr.Geometry:
    """
    Convert an OGR bounding box to ogr.Geometry.
    `[x_min, x_max, y_min, y_max] -> ogr.Geometry`

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    Returns
    -------
    ogr.Geometry
        An OGR geometry.
    """
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    ring = ogr.Geometry(ogr.wkbLinearRing)

    ring.AddPoint(x_min, y_min)
    ring.AddPoint(x_max, y_min)
    ring.AddPoint(x_max, y_max)
    ring.AddPoint(x_min, y_max)
    ring.AddPoint(x_min, y_min)

    geom = ogr.Geometry(ogr.wkbPolygon)
    geom.AddGeometry(ring)

    return geom


def _get_bbox_from_geom(geom: ogr.Geometry) -> List[Union[int, float]]:
    """
    Convert an ogr.Geometry to an OGR bounding box.
    `ogr.Geometry -> [x_min, x_max, y_min, y_max]`

    Parameters
    ----------
    geom : ogr.Geometry
        An OGR geometry.

    Returns
    -------
    list
        An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    assert isinstance(
        geom, ogr.Geometry
    ), f"geom was not a valid ogr.Geometry. Received: {geom}"

    bbox_ogr = list(geom.GetEnvelope()) # [x_min, x_max, y_min, y_max]

    return bbox_ogr


def _get_geotransform_from_bbox(
    bbox_ogr: List[Union[int, float]],
    raster_x_size: int,
    raster_y_size: int,
) -> List[Union[int, float]]:
    """
    Convert an OGR formatted bounding box to a GDAL GeoTransform.
    `[x_min, x_max, y_min, y_max] -> [x_min, pixel_width, x_skew, y_max, y_skew, pixel_height]`

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
    
    raster_x_size : int
        The number of pixels in the x direction.

    raster_y_size : int
        The number of pixels in the y direction.

    Returns
    -------
    list
        A GDAL GeoTransform. `[x_min, pixel_width, x_skew, y_max, y_skew, pixel_height]`
    """
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    origin_x = x_min
    origin_y = y_max
    pixel_width = (x_max - x_min) / raster_x_size
    pixel_height = (y_max - y_min) / raster_y_size

    return [origin_x, pixel_width, 0, origin_y, 0, utils_base._ensure_negative(pixel_height)]


def _get_gdal_bbox_from_ogr_bbox(
    bbox_ogr: List[Union[int, float]],
) -> List[Union[int, float]]:
    """
    Converts an OGR formatted bbox to a GDAL formatted one.
    `[x_min, x_max, y_min, y_max] -> [x_min, y_min, x_max, y_max]`

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.

    Returns
    -------
    list
        A GDAL formatted bbox. `[x_min, y_min, x_max, y_max]`
    """
    if isinstance(bbox_ogr, tuple):
        bbox_ogr = [val for val in bbox_ogr]

    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    return [x_min, y_min, x_max, y_max]


def _get_ogr_bbox_from_gdal_bbox(
    bbox_gdal: List[Union[int, float]],
) -> List[Union[int, float]]:
    """
    Converts a GDAL formatted bbox to an OGR formatted one.
    `[x_min, y_min, x_max, y_max] -> [x_min, x_max, y_min, y_max]`

    Parameters
    ----------
    bbox_gdal : list
        A GDAL formatted bbox.

    Returns
    -------
    list
        An OGR formatted bbox. `[x_min, x_max, y_min, y_max]`
    """
    if isinstance(bbox_ogr, tuple):
        bbox_ogr = [val for val in bbox_ogr]

    assert (
        isinstance(bbox_gdal, list) and len(bbox_gdal) == 4
    ), f"bbox_gdal must be a list of length four. Received: {bbox_gdal}."

    x_min, y_min, x_max, y_max = bbox_gdal

    return [x_min, x_max, y_min, y_max]


def _get_wkt_from_bbox(
    bbox_ogr: List[Union[int, float]],
) -> str:
    """
    Converts an OGR formatted bbox to a WKT string.
    `[x_min, x_max, y_min, y_max] -> WKT`

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.

    Returns
    -------
    str
        A WKT string. `POLYGON ((...))`
    """
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    wkt = f"POLYGON (({x_min} {y_min}, {x_max} {y_min}, {x_max} {y_max}, {x_min} {y_max}, {x_min} {y_min}))"

    return wkt


def _get_geojson_from_bbox(
    bbox_ogr: List[Union[int, float]],
) -> Dict[str, Union[str, List]]:
    """
    Converts an OGR formatted bbox to a GeoJson dictionary.
    `[x_min, x_max, y_min, y_max] -> GeoJson`

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
    
    Returns
    -------
    dict
        A GeoJson dictionary. `{ "type": "Polygon", "coordinates": [ ... ] }`
    """
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    geojson = {
        "type": "Polygon",
        "coordinates": [
            [
                [x_min, y_min],
                [x_max, y_min],
                [x_max, y_max],
                [x_min, y_max],
                [x_min, y_min],
            ]
        ],
    }

    return geojson


def _get_vector_from_bbox(
    bbox_ogr: List[Union[int, float]],
    projection_osr: osr.SpatialReference,
) -> ogr.DataSource:
    """
    Converts an OGR formatted bbox to an in-memory vector.
    _Vectors are stored in /vsimem/ as .gpkg files.
    **OBS**: Layers should be manually cleared when no longer used.

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
    
    projection_osr : osr.SpatialReference
        The projection of the vector.

    Returns
    -------
    ogr.DataSource
        The bounding box as a vector.
    """
    assert _check_is_valid_bbox(
        bbox_ogr
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr}"
    assert isinstance(
        projection_osr, osr.SpatialReference
    ), f"projection_osr not a valid spatial reference. Recieved: {projection_osr}"

    geom = _get_geom_from_bbox(bbox_ogr)

    driver = ogr.GetDriverByName("GPKG")
    extent_name = f"/vsimem/{str(uuid4().int)}_extent.gpkg"

    extent_ds = driver.CreateDataSource(extent_name)

    layer = extent_ds.CreateLayer("extent_ogr", projection_osr, ogr.wkbPolygon)

    feature = ogr.Feature(layer.GetLayerDefn())
    feature.SetGeometry(geom)
    layer.CreateFeature(feature)

    feature = None
    layer.SyncToDisk()

    return extent_name


def _check_bboxes_intersect(
    bbox1_ogr: List[Union[int, float]],
    bbox2_ogr: List[Union[int, float]],
) -> bool:
    """
    Checks if two OGR formatted bboxes intersect.

    Parameters
    ----------
    bbox1_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    bbox2_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    Returns
    -------
    bool
        **True** if the bboxes intersect, **False** otherwise.
    """
    assert _check_is_valid_bbox(
        bbox1_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox1_ogr}"
    assert _check_is_valid_bbox(
        bbox2_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox2_ogr}"

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    if bbox2_x_min > bbox1_x_max:
        return False

    if bbox2_y_min > bbox1_y_max:
        return False

    if bbox2_x_max < bbox1_x_min:
        return False

    if bbox2_y_max < bbox1_y_min:
        return False

    return True


def _check_bboxes_within(
    bbox1_ogr: List[Union[int, float]],
    bbox2_ogr: List[Union[int, float]],
) -> bool:
    """
    Checks if one OGR formatted bbox is within another.

    Parameters
    ----------
    bbox1_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    bbox2_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    Returns
    -------
    bool
        **True** if the bbox is within the other, **False** otherwise.
    """
    assert _check_is_valid_bbox(
        bbox1_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox1_ogr}"
    assert _check_is_valid_bbox(
        bbox2_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox2_ogr}"

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    return (
        (bbox1_x_min >= bbox2_x_min)
        and (bbox1_x_max <= bbox2_x_max)
        and (bbox1_y_min >= bbox2_y_min)
        and (bbox1_y_max <= bbox2_y_max)
    )


def _get_intersection_bboxes(
    bbox1_ogr: List[Union[int, float]],
    bbox2_ogr: List[Union[int, float]],
) -> List[Union[int, float]]:
    """
    Get the intersection of two OGR formatted bboxes.

    Parameters
    ----------
    bbox1_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
    
    bbox2_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
    
    Returns
    -------
    list
        An OGR formatted bbox of the intersection.
    """
    assert _check_bboxes_intersect(
        bbox1_ogr, bbox2_ogr
    ), "The two bounding boxes do not intersect."

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    return [
        max(bbox1_x_min, bbox2_x_min),
        min(bbox1_x_max, bbox2_x_max),
        max(bbox1_y_min, bbox2_y_min),
        min(bbox1_y_max, bbox2_y_max),
    ]


def _get_union_bboxes(
    bbox1_ogr: List[Union[int, float]],
    bbox2_ogr: List[Union[int, float]],
) -> List[Union[int, float]]:
    """
    Get the union of two OGR formatted bboxes.

    Parameters
    ----------
    bbox1_ogr : list
        An OGR formatted bbox.

    bbox2_ogr : list
        An OGR formatted bbox.

    Returns
    -------
    list
        An OGR formatted bbox of the union.
    """
    assert _check_is_valid_bbox(
        bbox1_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox1_ogr}"
    assert _check_is_valid_bbox(
        bbox2_ogr
    ), f"bbox1_ogr was not a valid bbox. Received: {bbox2_ogr}"

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    return [
        min(bbox1_x_min, bbox2_x_min),
        max(bbox1_x_max, bbox2_x_max),
        min(bbox1_y_min, bbox2_y_min),
        max(bbox1_y_max, bbox2_y_max),
    ]


def _get_aligned_bbox_to_pixel_size(
    bbox1_ogr: List[Union[int, float]],
    bbox2_ogr: List[Union[int, float]],
    pixel_width: Union[int, float],
    pixel_height: Union[int, float],
) -> List[Union[int, float]]:
    """
    Aligns two OGR formatted bboxes to a pixel size. Output is an augmented version
    of bbox2.

    Parameters
    ----------
    bbox1_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`
    
    bbox2_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    pixel_width : int or float
        The width of the pixel.
    
    pixel_height : int or float
        The height of the pixel.

    Returns
    -------
    list
        An OGR formatted bbox of the alignment.
        `[x_min, x_max, y_min, y_max]`
    """
    assert (
        isinstance(bbox1_ogr, list) and len(bbox1_ogr) == 4
    ), f"bbox1_ogr must be a list of length four. Received: {bbox1_ogr}."
    assert (
        isinstance(bbox2_ogr, list) and len(bbox2_ogr) == 4
    ), f"bbox2_ogr must be a list of length four. Received: {bbox2_ogr}."

    bbox1_x_min, bbox1_x_max, bbox1_y_min, bbox1_y_max = bbox1_ogr
    bbox2_x_min, bbox2_x_max, bbox2_y_min, bbox2_y_max = bbox2_ogr

    x_min = bbox2_x_min - ((bbox2_x_min - bbox1_x_min) % pixel_width)
    x_max = bbox2_x_max + ((bbox1_x_max - bbox2_x_max) % pixel_width)

    y_min = bbox2_y_min - ((bbox2_y_min - bbox1_y_min) % abs(pixel_height))
    y_max = bbox2_y_max + ((bbox1_y_max - bbox2_y_max) % abs(pixel_height))

    return x_min, x_max, y_min, y_max


def _get_bounds_from_bbox(
    bbox_ogr: List[Union[int, float]],
    projection_osr: osr.SpatialReference,
    wkt: bool = True,
) -> ogr.Geometry:
    """ 
    This is an internal utility function for metadata generation. It takes a standard
    OGR bounding box and the geometry of the source dataset in latlng and returns a
    geometry in the source dataset's projection. This is important as, as when 
    reprojecting to latlng, you might get a skewed bounding box. This function returns
    the skewed bounds in latlng, which is what you want for overlap checks across
    projections.

    The return is in WGS84.

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    projection_osr : osr.SpatialReference
        The projection of the bbox

    wkt : bool, optional
        Whether to return the geometry as a WKT string or an OGR geometry object.

    Returns
    -------
    ogr.Geometry
        The geometry in the source dataset's projection.
    """
    default_projection = utils_projection._get_default_projection_osr()

    if utils_projection._check_projections_match(projection_osr, default_projection):
        geom = _get_geom_from_bbox(bbox_ogr)

        if wkt:
            return geom.ExportToWkt()

        return geom

    options = osr.CoordinateTransformationOptions()
    options.SetAreaOfInterest(-180.0, -90.0, 180.0, 90.0)
    transformer = osr.CoordinateTransformation(projection_osr, default_projection, options)

    x_min, x_max, y_min, y_max = bbox_ogr

    p1 = [x_min, y_min]
    p2 = [x_max, y_min]
    p3 = [x_max, y_max]
    p4 = [x_min, y_max]

    gdal.PushErrorHandler("CPLQuietErrorHandler")
    try:
        p1t = transformer.TransformPoint(p1[0], p1[1])
        p2t = transformer.TransformPoint(p2[0], p2[1])
        p3t = transformer.TransformPoint(p3[0], p3[1])
        p4t = transformer.TransformPoint(p4[0], p4[1])
    except RuntimeError:
        p1t = transformer.TransformPoint(float(p1[0]), float(p1[1]))
        p2t = transformer.TransformPoint(float(p2[0]), float(p2[1]))
        p3t = transformer.TransformPoint(float(p3[0]), float(p3[1]))
        p4t = transformer.TransformPoint(float(p4[0]), float(p4[1]))

    gdal.PopErrorHandler()

    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(p1t[1], p1t[0])
    ring.AddPoint(p2t[1], p2t[0])
    ring.AddPoint(p3t[1], p3t[0])
    ring.AddPoint(p4t[1], p4t[0])
    ring.AddPoint(p1t[1], p1t[0])

    polygon = ogr.Geometry(ogr.wkbPolygon)
    polygon.AddGeometry(ring)

    if wkt:
        return polygon.ExportToWkt()

    return polygon


def _get_utm_zone_from_bbox(
    bbox_ogr_latlng: List[Union[int, float]],
) -> osr.SpatialReference:
    """
    Get the UTM zone from an OGR formatted bbox.

    Parameters
    ----------
    bbox_ogr_latlng : list
        An OGR formatted bbox.

    Returns
    -------
    osr.SpatialReference
        The UTM zone projection.
    """
    assert _check_is_valid_bbox(
        bbox_ogr_latlng
    ), f"bbox_ogr was not a valid bbox. Received: {bbox_ogr_latlng}"
    assert _check_is_valid_bbox_latlng(bbox_ogr_latlng), "Bbox is not in latlng format."

    bbox_x_min, bbox_x_max, bbox_y_min, bbox_y_max = bbox_ogr_latlng

    mid_lng = (bbox_x_min + bbox_x_max) / 2
    mid_lat = (bbox_y_min + bbox_y_max) / 2

    return utils_projection._get_utm_zone_from_latlng([mid_lat, mid_lng])


def _get_utm_zone_from_dataset(
    dataset: Union[str, gdal.Dataset, ogr.DataSource],
) -> str:
    """
    Get the UTM zone from a GDAL dataset.

    Parameters
    ----------
    dataset : str or gdal.Dataset or ogr.DataSource
        A GDAL dataset.

    Returns
    -------
    str
        The UTM zone.
    """
    assert isinstance(dataset, (str, gdal.Dataset, ogr.DataSource)), "dataset was not a valid dataset."

    bbox = get_bbox_from_dataset(dataset)
    source_projection = utils_projection._get_projection_from_dataset(dataset)
    target_projection = osr.SpatialReference()
    target_projection_wkt = utils_projection._get_default_projection()
    target_projection.ImportFromWkt(target_projection_wkt)

    bbox = utils_projection.reproject_bbox(bbox, source_projection, target_projection)
    utm_zone = _get_utm_zone_from_bbox(bbox)

    return utm_zone


def _get_utm_zone_from_dataset_list(
    datasets: List[Union[str, gdal.Dataset, ogr.DataSource]],
) -> str:
    """
    Get the UTM zone from a list of GDAL datasets.

    Parameters
    ----------
    datasets : list
        A list of GDAL datasets.

    Returns
    -------
    str
        The UTM zone.
    """
    assert isinstance(datasets, (list, np.ndarray)), "datasets was not a valid list."

    latlng_bboxes = []
    latlng_proj = utils_projection._get_default_projection_osr()

    for dataset in datasets:
        bbox = get_bbox_from_dataset(dataset)
        projection = utils_projection._get_projection_from_dataset(dataset)
        latlng_bboxes.append(utils_projection.reproject_bbox(bbox, projection, latlng_proj))

    union_bbox = latlng_bboxes[0]
    for bbox in latlng_bboxes[1:]:
        union_bbox = _get_union_bboxes(union_bbox, bbox)

    utm_zone = _get_utm_zone_from_bbox(union_bbox)

    return utm_zone


def _additional_bboxes(
    bbox_ogr: List[Union[int, float]],
    projection_osr: osr.SpatialReference,
) -> Dict[str, Any]:
    """
    This is an internal utility function for metadata generation. It takes a standard
    OGR bounding box and returns a list of variations of bounding boxes.

    Parameters
    ----------
    bbox_ogr : list
        An OGR formatted bbox.
        `[x_min, x_max, y_min, y_max]`

    projection_osr : osr.SpatialReference
        The projection.

    Returns
    -------
    dict
        A dictionary of the added bboxes. Contains the following keys:
        `bbox_latlng`: The bbox in latlng coordinates.
        `bbox_wkt`: The bbox in WKT format.
        `bbox_wkt_latlng`: The bbox in WKT format in latlng coordinates.
        `bbox_geom`: The bbox in ogr.Geometry format.
        `bbox_geom_latlng`: The bbox in ogr.Geometry format in latlng coordinates.
        `bbox_gdal`: The bbox in GDAL format.
        `bbox_gdal_latlng`: The bbox in GDAL format in latlng coordinates.
        `bbox_dict`: The bbox in a dictionary format.
        `bbox_dict_latlng`: The bbox in a dictionary format in latlng coordinates.
    """
    assert _check_is_valid_bbox(bbox_ogr), f"Invalid bbox. Received: {bbox_ogr}."
    assert isinstance(
        projection_osr, osr.SpatialReference
    ), f"source_projection not a valid spatial reference. Recieved: {projection_osr}"

    x_min, x_max, y_min, y_max = bbox_ogr

    original_projection = osr.SpatialReference()
    original_projection.ImportFromWkt(projection_osr.ExportToWkt())

    latlng_projection = utils_projection._get_default_projection_osr()
    bbox_ogr_latlng = utils_projection.reproject_bbox(bbox_ogr, original_projection, latlng_projection)

    world = False
    if np.isinf(bbox_ogr_latlng).any():
        world = True
    if np.isinf(bbox_ogr_latlng[0]):
        bbox_ogr_latlng[0] = -179.999999
    if np.isinf(bbox_ogr_latlng[1]):
        bbox_ogr_latlng[1] = 180.0
    if np.isinf(bbox_ogr_latlng[2]):
        bbox_ogr_latlng[2] = -89.999999
    if np.isinf(bbox_ogr_latlng[3]):
        bbox_ogr_latlng[3] = 90.0

    latlng_x_min, latlng_x_max, latlng_y_min, latlng_y_max = bbox_ogr_latlng

    bbox_geom = _get_geom_from_bbox(bbox_ogr)
    bbox_geom_latlng = _get_geom_from_bbox(bbox_ogr_latlng)
    geom = bbox_geom

    if world:
        geom_latlng = bbox_geom_latlng
    else:
        geom_latlng = _get_bounds_from_bbox(bbox_ogr, original_projection, latlng_projection)

    geom_latlng_geom = ogr.CreateGeometryFromWkt(geom_latlng)

    bbox_wkt = _get_wkt_from_bbox(bbox_ogr)
    bbox_wkt_latlng = _get_wkt_from_bbox(bbox_ogr_latlng)

    return {
        "bbox_latlng": bbox_ogr_latlng,
        "bbox_wkt": bbox_wkt,
        "bbox_wkt_latlng": bbox_wkt_latlng,
        "bbox_geom": bbox_geom,
        "bbox_geom_latlng": bbox_geom_latlng,
        "bbox_gdal": _get_gdal_bbox_from_ogr_bbox(bbox_ogr),
        "bbox_gdal_latlng": _get_gdal_bbox_from_ogr_bbox(bbox_ogr_latlng),
        "bbox_dict": {"x_min": x_min, "x_max": x_max, "y_min": y_min, "y_max": y_max},
        "bbox_dict_latlng": {
            "x_min": latlng_x_min,
            "x_max": latlng_x_max,
            "y_min": latlng_y_min,
            "y_max": latlng_y_max,
        },
        "bbox_geojson": _get_geojson_from_bbox(bbox_ogr_latlng),
        "area_latlng": geom_latlng_geom.GetArea(),
        "area": geom.GetArea(),
        "geom": geom,
        "geom_latlng": geom_latlng,
    }


def _get_vector_from_geom(
    geom: ogr.Geometry,
    out_path: Optional[str] = None,
    name: Optional[str] = "converted_geom",
    prefix: Optional[str] = "",
    suffix: Optional[str] = "",
    add_uuid: Optional[bool] = False,
    add_timestamp: Optional[bool] = False,
    projection_osr: Optional[osr.SpatialReference] = None,
) -> ogr.DataSource:
    """
    Converts a geometry to a vector in memory.

    Parameters
    ----------
    geom : ogr.Geometry
        The geometry to convert.

    Returns
    -------
    ogr.DataSource
    """
    assert isinstance(geom, ogr.Geometry), "geom must be an ogr.Geometry."

    if out_path is None:
        path = utils_path._get_temp_filepath(
            f"{name}.gpkg",
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
            add_timestamp=add_timestamp,
        )
    else:
        assert isinstance(out_path, str), "out_path must be a string."
        assert utils_path._check_is_valid_output_filepath(out_path), "out_path was not a valid output filepath."

        path = out_path

    driver_name = utils_gdal._get_vector_driver_name_from_path(path)
    driver = ogr.GetDriverByName(driver_name)
    vector = driver.CreateDataSource(path)

    proj = projection_osr if projection_osr is not None else utils_projection._get_default_projection_osr()
    layer = vector.CreateLayer(name, proj, geom.GetGeometryType())

    feature = ogr.Feature(layer.GetLayerDefn())
    feature.SetGeometry(geom)

    layer.CreateFeature(feature)
    feature.Destroy()

    vector = None

    return path
